@@ -16,26 +16,24 @@
 //
 //
 
-#include <grpc/support/port_platform.h>
-
 #include "src/core/ext/transport/chttp2/transport/frame_ping.h"
 
+#include <grpc/support/alloc.h>
+#include <grpc/support/port_platform.h>
+#include <inttypes.h>
 #include <string.h>
 
 #include <algorithm>
-#include <initializer_list>
 
+#include "absl/container/flat_hash_map.h"
+#include "absl/log/check.h"
+#include "absl/log/log.h"
 #include "absl/status/status.h"
 #include "absl/strings/str_format.h"
-
-#include <grpc/support/alloc.h>
-#include <grpc/support/log.h>
-
 #include "src/core/ext/transport/chttp2/transport/internal.h"
-#include "src/core/ext/transport/chttp2/transport/stream_map.h"
-#include "src/core/lib/gprpp/time.h"
-
-static bool g_disable_ping_ack = false;
+#include "src/core/ext/transport/chttp2/transport/ping_abuse_policy.h"
+#include "src/core/ext/transport/chttp2/transport/ping_callbacks.h"
+#include "src/core/lib/debug/trace.h"
 
 grpc_slice grpc_chttp2_ping_create(uint8_t ack, uint64_t opaque_8bytes) {
   grpc_slice slice = GRPC_SLICE_MALLOC(9 + 8);
@@ -91,32 +89,32 @@ grpc_error_handle grpc_chttp2_ping_parser_parse(void* parser,
   }
 
   if (p->byte == 8) {
-    GPR_ASSERT(is_last);
+    CHECK(is_last);
+    t->http2_ztrace_collector.Append(
+        grpc_core::H2PingTrace<true>{p->is_ack != 0, p->opaque_8bytes});
     if (p->is_ack) {
+      GRPC_TRACE_LOG(http2_ping, INFO)
+          << (t->is_client ? "CLIENT" : "SERVER") << "[" << t
+          << "]: received ping ack " << p->opaque_8bytes;
       grpc_chttp2_ack_ping(t, p->opaque_8bytes);
     } else {
       if (!t->is_client) {
-        grpc_core::Timestamp now = grpc_core::Timestamp::Now();
-        grpc_core::Timestamp next_allowed_ping =
-            t->ping_recv_state.last_ping_recv_time +
-            t->ping_policy.min_recv_ping_interval_without_data;
-
-        if (t->keepalive_permit_without_calls == 0 &&
-            grpc_chttp2_stream_map_size(&t->stream_map) == 0) {
-          // According to RFC1122, the interval of TCP Keep-Alive is default to
-          // no less than two hours. When there is no outstanding streams, we
-          // restrict the number of PINGS equivalent to TCP Keep-Alive.
-          next_allowed_ping = t->ping_recv_state.last_ping_recv_time +
-                              grpc_core::Duration::Hours(2);
+        const bool transport_idle =
+            t->keepalive_permit_without_calls == 0 && t->stream_map.empty();
+        if (GRPC_TRACE_FLAG_ENABLED(http_keepalive) ||
+            GRPC_TRACE_FLAG_ENABLED(http)) {
+          LOG(INFO) << "SERVER[" << t << "]: received ping " << p->opaque_8bytes
+                    << ": "
+                    << t->ping_abuse_policy.GetDebugString(transport_idle);
         }
-
-        if (next_allowed_ping > now) {
-          grpc_chttp2_add_ping_strike(t);
+        if (t->ping_abuse_policy.ReceivedOnePing(transport_idle)) {
+          grpc_chttp2_exceeded_ping_strikes(t);
         }
-
-        t->ping_recv_state.last_ping_recv_time = now;
+      } else {
+        GRPC_TRACE_LOG(http2_ping, INFO)
+            << "CLIENT[" << t << "]: received ping " << p->opaque_8bytes;
       }
-      if (!g_disable_ping_ack) {
+      if (t->ack_pings) {
         if (t->ping_ack_count == t->ping_ack_capacity) {
           t->ping_ack_capacity =
               std::max(t->ping_ack_capacity * 3 / 2, size_t{3});
@@ -132,7 +130,3 @@ grpc_error_handle grpc_chttp2_ping_parser_parse(void* parser,
 
   return absl::OkStatus();
 }
-
-void grpc_set_disable_ping_ack(bool disable_ping_ack) {
-  g_disable_ping_ack = disable_ping_ack;
-}
